	; -*-asm-*-    set emacs asm-mode

	.title MSP430eForth
	.psize 62,132	;62 lines per page, 132 characters per line
	.equ	hardwareUM,1	; enable use of hardware multiplier

;===============================================================
;
;       MSP430 eForth 2.02 D. W. Schultz, 2012
;       To be assembled by the GNU assembler.
;
;	Port to MSP430FR5969 Launchpad 2/24/2015
;         See system.S for machine specific details
;
; 	Added a few useful words: ['],<BUILDS,DOES>
;	
;	
;	
;	
;	Based on:
;
;       86eForth 2.02, C. H. Ting, 06/02/99
;       Add create, checksum, UPLOAD and DOWNLOAD.
;       A sample session looks like:
;               c>86ef202
;               DOWNLOAD LESSONS.TXT
;               WORDS
;               ' THEORY 'BOOT !
;               UPLOAD TEST.EXE
;               BYE
;               c>test
;
;       86eForth 2.01, C. H. Ting, 05/24/99
;       Merge Zen2.asm with eForth 1.12
;1.     Eliminate most of the @EXECUTE thru user variables
;2.     Combine name and code dictionary
;3.     Eliminate code pointer fields
;4.     elimiate catch-throw
;5.     eliminate most user variables
;6.     extend top memory to FFF0H where the stacks and user area are.
;7.     add open, close, read, write; improve BYE
;8      add 1+, 1-, 2/
;
;       
;       eForth 1.12, C. H. Ting, 03/30/99
;               Change READ and LOAD to 'read' and 'load'.
;               Make LOAD to read and compile a file.  The file
;               buffer is from CP+1000 to NP-100.
;               To load all the lessons, type:
;                       LOAD LESSONS.TXT
;               and you can test all the examples in this file.
;       eForth 1.11, C. H. Ting, 03/25/99
;               Change BYE to use function 4CH of INT 21H.
;               Add read, write, open, close, READ, and LOAD
;               To read a text file into memory:
;                       HEX 2000 1000 READ TEST.TXT
;               READ returns the number of byte actually read.
;               To compile the source code in the text file:
;                       2000 FCD LOAD
;               where FCD is the length returned by READ.
;               These additions allow code for other eForth systems
;               to be tested on PC first.
;               It is part of the Firmware Engineering Workshop.
;
;
;	eForth 1.0 by Bill Muench and C. H. Ting, 1990
;	Much of the code is derived from the following sources:
;		8086 figForth by Thomas Newman, 1981 and Joe smith, 1983
;		aFORTH by John Rible
;		bFORTH by Bill Muench
;
;	The goal of this implementation is to provide a simple eForth Model
;	which can be ported easily to many 8, 16, 24 and 32 bit CPU's.
;	The following attributes make it suitable for CPU's of the '90:
;
;		small machine dependent kernel and portable high level code
;		source code in the MASM format
;		direct threaded code
;		separated code and name dictionaries
;		simple vectored terminal and file interface to host computer
;		aligned with the proposed ANS Forth Standard
;		easy upgrade path to optimize for specific CPU
;
;	You are invited to implement this Model on your favorite CPU and
;	contribute it to the eForth Library for public use. You may use
;	a portable implementation to advertise more sophisticated and
;	optimized version for commercial purposes. However, you are
;	expected to implement the Model faithfully. The eForth Working
;	Group reserves the right to reject implementation which deviates
;	significantly from this Model.
;
;	As the ANS Forth Standard is still evolving, this Model will
;	change accordingly. Implementations must state clearly the
;	version number of the Model being tracked.
;
;	Representing the eForth Working Group in the Silicon Valley FIG Chapter.
;	Send contributions to:
;
;		Dr. C. H. Ting
;		156 14th Avenue
;		San Mateo, CA 94402
;		(415) 571-7639
;
;===============================================================
;; Version control

	.equ VER, 2                       ;major release version
	.equ EXT, 2                       ;minor extension

;; Constants

	.equ TRUEE,-1		;true flag

	.equ COMPO,0x040		;lexicon compile only bit
	.equ IMEDD,0x080		;lexicon immediate bit
	.equ MASKK,0x07F1F		;lexicon bit mask
	
	.equ CELLL,2		;size of a cell
	.equ BASEE,10		;default radix
	.equ VOCSS,8		;depth of vocabulary stack

	.equ BKSPP,	8		;back space
	.equ LF,	10		;line feed
	.equ CRR,	13		;carriage return
	.equ ERR,	27		;error escape
	.equ TIC,	39		;tick

	.equ CALLL,	0x1287		; CALL doenter (call r7)
	.equ CALLD,     0x1288		; CALL dodoes  (call r8)
	
;; Memory allocation

	.equ EM, __stack	;top of memory
	.equ US, 64*CELLL	;user area size in cells
	.equ RTS,64*CELLL	;return stack/TIB size

	.equ TIBB, EM-RTS	;terminal input buffer (TIB) 
	.equ UPP,TIBB-RTS	;start of user area (UP0)
	.equ RPP,UPP		;start of return stack (RP0)
	.equ SPP, UPP-RTS	;start of data stack (SP0)

;; FORTH virtual machine - register use
;; Register r9-r15 are available for use in code words
;; 
	.equ	PSP,r1		; parameter stack pointer (also hardware stack)
	.equ	RSP,r4		; return stack pointer
	.equ	IPTR,r5		; inner interpreter pointer
	.equ	TOS,r6		; Top Of Stack
	.equ	DOCOL,r7	; address of doenter routine (for call r7)
	.equ	DODOES,r8	; address of dodoes routine (for call r8)
	
;; Define assembly macros


	.macro   CODE  lex, name, label
	.word  _link
	.equ _link, .
	.byte  \lex
	.ascii   "\name"
	.align 1                ; force 16 bit alignment
\label :
	.endm

;; Compile a colon definition header
	.macro  COLON  lex, name, label
	CODE \lex, "\name", \label
	call	DOCOL
	.endm

;	Compile a user variable header.
	.macro USER  lex,name,label
	CODE \lex, "\name", \label
	call	DOCOL
	.word  DOUSE, _user
	.equ _user, _user+CELLL
	.endm

;	Assemble inline direct threaded code ending.

	.macro	next
	mov	@IPTR+,r0
	.endm

;; Main entry points and COLD start data
	.text
	.globl	main
	.func
	main:			; Init various registers...
	mov	#RPP,RSP	 ;  return stack pointer
	mov	#SPP,PSP	 ; parameter stack
	mov	#doenter,DOCOL	 ;  do colon pointer
	mov	#dodoes,DODOES	 ; and dodoes
	mov	#COLD1,IPTR	 ; instruction pointer
	next
	.endfunc	
	
; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.
UZERO:
	.word	BASEE		;BASE
	.word	0		;tmp
	.word	0		;>IN
	.word	0		;#TIB
	.word	TIBB		;TIB
	.word	INTER		;'EVAL
	.word	0		;HLD
	.word	0		;CONTEXT pointer
	.word	CTOP		;CP
	.word   LASTN           ;LAST
ULAST:  .word   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0

	;; ULAST is used by system.S to find CTOP which it needs to
	;; init the memory protection unit
	.globl	ULAST
		
;; Initialize assembly variables

	.equ _link, 0	;force a null link
	.equ _user, 0           ;first user variable offset

;; Device dependent I/O

;   ?RX		( -- c T | F )
;		Return input character and true, or a false if no input.

	        CODE   4,"?KEY",QKEY
	push	TOS
	call	#rx
	tst	r15	; check flag
	jeq	no_key
	push	r14	; push char
no_key:	mov	r15,TOS	; move flag to TOS
	next
;   TX!		( c -- )
;		Send character c to the output device.

	CODE   4,"EMIT",EMIT
	mov	TOS,r15
	call 	#tx
	pop	TOS; pop new TOS
	next

;; The kernel

;; Start a COLON definition.
;; Invoked with "call DOCOL" which pushes what will be the new IP
;; to the parameter stack.
;; 
doenter:
	decd	RSP
	mov	IPTR,@RSP; push return address
	mov	@PSP+,IPTR; pop new instruction pointer
	next
dodoes:
	decd	RSP
	mov	IPTR,@RSP	; push return address
	mov	@PSP+,r15	; get parameter address
	push	TOS		; make room for new TOS
	mov	@r15+,IPTR	; pop new instruction pointer
	mov	r15,TOS		; address of parameter field
	next
	
	
;   doLIT	( -- w )
;		Push an inline literal.

	CODE   COMPO+5,"doLit",DOLIT
	push	TOS
	mov	@IPTR+,TOS
	next

;   next	( -- )
;		Run time code for the single index loop.
;		: next ( -- ) \ hilevel model
;		  r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

	CODE   COMPO+4,"next",DONXT
	mov	@RSP,r15	;decrement index
	dec	r15
	mov	r15,@RSP
	jlo	NEXT1		; decrement below zero ?
	mov	@IPTR,IPTR	; no, branch
	next
NEXT1:	incd	RSP		; drop index counter from return stack
	incd	IPTR		; skip branch target
	next

;   ?branch	( f -- )
;		Branch if flag is zero.

	CODE   COMPO+7,"?branch",QBRAN
	tst	TOS	; test TOS
	pop	TOS	; pop new TOS
	jeq	BRAN1	; TOS zero?
	incd	IPTR	; skip branch target
	next
BRAN1:	mov	@IPTR,IPTR; take branch by loading new IP
	next	

;   branch	( -- )
;		Branch to an inline address.

	CODE   COMPO+6,"branch",BRAN
	mov	@IPTR,IPTR
	next

;   EXECUTE	( ca -- )
;		Execute the word at ca.

	CODE	7,"EXECUTE",EXECU
	mov	TOS, r15
	pop	TOS; get new TOS
	mov	r15,r0

;   EXIT	( -- )
;		Terminate a colon definition.

	CODE	4,"EXIT",EXIT
	mov	@RSP+,IPTR; pop return address
	next

;   !		( w a -- )
;		Pop the data stack to memory.

	CODE	1,"!",STORE
	pop	r15	; get data
	mov	r15,@TOS; store data
	pop	TOS	; get new TOS
	next

;   @		( a -- w )
;		Push memory location to the data stack.

	CODE	1,"@",AT
	mov	@TOS,TOS
	next

;   C!		( c b -- )
;		Pop the data stack to byte memory.

	CODE	2,"C!",CSTOR
	pop	r15
	mov.b	r15,@TOS; store data
	pop	TOS	; get new TOS
	next

;   C@		( b -- c )
;		Push byte memory location to the data stack.

	CODE	2,"C@",CAT
	mov.b	@TOS,TOS
	and.w	#0xff,TOS
	next

	;; Basic support for accessing high memory on CPUX devices
	;; The natural order for storing data on the MSP430 is different
	;; from the standard Forth order where the high word would be
	;; at the lower address. Maintaining consistency requires
	;; more work to shuffle this around. If there is an easier
	;; way, I can't think of it.
	;; 
; X@    (a a -- w)
; Using the double word address, pop the data stack to memory

	CODE	2,"X@",XAT
	pop	r15		; get low half of address
	push	TOS		; push high half of address to  memory
	push	r15		; push low half
	popx.a	r15		; get 20 bit address into a register
	mov	@r15,TOS	; fetch data
	next
	
; X!    (a a w -- )
; Using the double word address, push the memory location to the data stack

	CODE	2,"X!",XSTORE
	pop	r15		; get high half of address
	pop	r14		; and low half
	push	r15		; push to stack in correct order
	push	r14
	popx.a	r15		; get 20 bit address into a register
	mov	TOS,@r15	; write data
	pop	TOS		; replace TOS
	next
	
;   RP@		( -- a )
;		Push the current RP to the data stack.

	CODE   3,"rp@",RPAT
	push	TOS
	mov	RSP,TOS
	next

;   RP!		( a -- )
;		Set the return stack pointer.

	CODE   COMPO+3,"rp!",RPSTO
	mov	TOS,RSP
	pop	TOS
	next

;   R>		( -- w )
;		Pop the return stack to the data stack.

	CODE	COMPO+2,"R>",RFROM
	push	TOS	; push current TOS
	mov	@RSP+,TOS ; pop return stack to TOS
	next

;   R@		( -- w )
;		Copy top of return stack to the data stack.

	CODE	2,"R@",RAT
	push	TOS	; push current TOS
	mov	@RSP,TOS; copy top of return stack to TOS
	next

;   >R		( w -- )
;		Push the data stack to the return stack.

	CODE	COMPO+2,">R",TOR
	decd	RSP
	mov	TOS,@RSP; push TOS to return stack
	pop	TOS	; get new TOS
	next

;   SP@		( -- a )
;		Push the current data stack pointer.

	CODE   3,"sp@",SPAT
	push	TOS
	mov	PSP,TOS
	next

;   SP!		( a -- )
;		Set the data stack pointer.

	CODE   3,"sp!",SPSTO
	mov	TOS,PSP
	next

;   DROP	( w -- )
;		Discard top stack item.

	CODE	4,"DROP",DROP
	pop	TOS
	next

;   DUP		( w -- w w )
;		Duplicate the top stack item.

	CODE	3,"DUP",DUPP
	push	TOS
	next

;   SWAP	( w1 w2 -- w2 w1 )
;		Exchange top two stack items.

	CODE	4,"SWAP",SWAP
	mov	@PSP,r15
	mov	TOS,@PSP
	mov	r15,TOS
	next

;   OVER	( w1 w2 -- w1 w2 w1 )
;		Copy second stack item to top.

	CODE	4,"OVER",OVER
	push	TOS
	mov	2(PSP),TOS
	next

;   0<		( n -- t )
;		Return true if n is negative.

	CODE	2,"0<",ZLESS
	swpb	TOS	; kind of clunky but the alternative
	sxt	TOS	; is a test and jump
	swpb	TOS
	sxt	TOS
	next

;   AND		( w w -- w )
;		Bitwise AND.

	CODE	3,"AND",ANDD
	pop	r15
	and	r15,TOS
	next

;   OR		( w w -- w )
;		Bitwise inclusive OR.

	CODE	2,"OR",ORR
	pop	r15
	bis	r15,TOS
	next

;   XOR		( w w -- w )
;		Bitwise exclusive OR.

	CODE	3,"XOR",XORR
	pop	r15
	xor	r15,TOS
	next

;   UM+		( u u -- udsum )
;		Add two unsigned single numbers and return a double sum.

	CODE   3,"UM+",UPLUS
	clr	r15
	mov	@PSP,r14
	add	r14,TOS
	addc	r15,r15
	mov	TOS,@PSP
	mov	r15,TOS
	next

;; System and user variables

;   doVAR	( -- a )
;		Run time routine for VARIABLE and CREATE.
	
	COLON  COMPO+5,"doVar",DOVAR
	.word	RFROM,EXIT

;   UP		( -- a )
;		Pointer to the user area.

	COLON  2,"up",UP
	.word	DOVAR
	.word	UPP

;   doUSER	( -- a )
;		Run time routine for user variables.

	COLON  COMPO+6,"doUser",DOUSE
	.word	RFROM,AT,UP,AT,PLUS,EXIT

;   BASE	( -- a )
;		Storage of the radix base for numeric I/O.

	USER	4,"BASE",BASE

;   tmp		( -- a )
;		A temporary storage location used in parse and find.

	USER   COMPO+3,"tmp",TEMP

;   >IN		( -- a )
;		Hold the character pointer while parsing input stream.

	USER	3,">IN",INN

;   #TIB	( -- a )
; Hold the current count in and address of the terminal input buffer.

	USER	4,"#TIB",NTIB
	_user = _user+CELLL;hold the base address of the terminal input buffer

;   "EVAL	( -- a )
;		Execution vector of EVAL.

	USER   5,"\"eval",TEVAL

;   HLD		( -- a )
;		Hold a pointer in building a numeric output string.

	USER   3,"hld",HLD

;   CONTEXT	( -- a )
;		A area to specify vocabulary search order.

	USER   7,"CONTEXT",CNTXT

;   CP		( -- a )
;		Point to the top of the code dictionary.

	USER   2,"cp",CP

;   LAST	( -- a )
;		Point to the last name in the name dictionary.

	USER   4,"last",LAST

;; Common functions

;   ?DUP	( w -- w w | 0 )
;		Dup tos if its is not zero.

	COLON	4,"?DUP",QDUP
	.word	DUPP
	.word	QBRAN,QDUP1
	.word	DUPP
QDUP1:	.word	EXIT

;   ROT		( w1 w2 w3 -- w2 w3 w1 )
;		Rot 3rd item to top.

	COLON	3,"ROT",ROT
	.word	TOR,SWAP,RFROM,SWAP,EXIT

;   2DROP	( w w -- )
;		Discard two items on stack.

	COLON	5,"2DROP",DDROP
	.word	DROP,DROP,EXIT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;		Duplicate top two items.

	COLON	4,"2DUP",DDUP
	.word	OVER,OVER,EXIT

;   +		( w w -- sum )
;		Add top two items.

	COLON	1,"+",PLUS
	.word	UPLUS,DROP,EXIT

;   NOT		( w -- w )
;		One"s complement of tos.

	COLON	3,"NOT",INVER
	.word	DOLIT,-1,XORR,EXIT

;   NEGATE	( n -- -n )
;		Two"s complement of tos.

	COLON	6,"NEGATE",NEGAT
	.word      INVER,ONEP,EXIT

;   DNEGATE	( d -- -d )
;		Two"s complement of top double.

	COLON	7,"DNEGATE",DNEGA
	.word	INVER,TOR,INVER
	.word	DOLIT,1,UPLUS
	.word	RFROM,PLUS,EXIT

;   -		( n1 n2 -- n1-n2 )
;		Subtraction.

	COLON	1,"-",SUBB
	.word	NEGAT,PLUS,EXIT

;   ABS		( n -- n )
;		Return the absolute value of n.

	COLON	3,"ABS",ABSS
	.word	DUPP,ZLESS
	.word	QBRAN,ABS1
	.word	NEGAT
ABS1:	.word	EXIT

;   =		( w w -- t )
;		Return true if top two are equal.

	COLON	1,"=",EQUAL
	.word	XORR
	.word	QBRAN,EQU1
	.word	DOLIT,0,EXIT
EQU1:	.word	DOLIT,TRUEE,EXIT

;   U<		( u u -- t )
;		Unsigned compare of top two items.

	COLON	2,"U<",ULESS
	.word	DDUP,XORR,ZLESS
	.word	QBRAN,ULES1
	.word	SWAP,DROP,ZLESS,EXIT
ULES1:	.word	SUBB,ZLESS,EXIT

;   <		( n1 n2 -- t )
;		Signed compare of top two items.

	COLON	1,"<",LESS
	.word	DDUP,XORR,ZLESS
	.word	QBRAN,LESS1
	.word	DROP,ZLESS,EXIT
LESS1:	.word	SUBB,ZLESS,EXIT

;   MAX		( n n -- n )
;		Return the greater of two top stack items.

	COLON	3,"MAX",MAX
	.word	DDUP,LESS
	.word	QBRAN,MAX1
	.word	SWAP
MAX1:	.word	DROP,EXIT

;   MIN		( n n -- n )
;		Return the smaller of top two stack items.

	COLON	3,"MIN",MIN
	.word	DDUP,SWAP,LESS
	.word	QBRAN,MIN1
	.word	SWAP
MIN1:	.word	DROP,EXIT

;   WITHIN	( u ul uh -- t )
;		Return true if u is within the range of ul and uh. ( ul <= u < uh )

	COLON	6,"WITHIN",WITHI
	.word	OVER,SUBB,TOR
	.word	SUBB,RFROM,ULESS,EXIT

;; Divide

;   UM/MOD	( udl udh un -- ur uq )
;		Unsigned divide of a double by a single. Return mod and quotient.

	COLON	6,"UM/MOD",UMMOD
	.word	DDUP,ULESS
	.word	QBRAN,UMM4
	.word	NEGAT,DOLIT,15,TOR
UMM1:	.word	TOR,DUPP,UPLUS
	.word	TOR,TOR,DUPP,UPLUS
	.word	RFROM,PLUS,DUPP
	.word	RFROM,RAT,SWAP,TOR
	.word	UPLUS,RFROM,ORR
	.word	QBRAN,UMM2
	.word      TOR,DROP,ONEP,RFROM
	.word	BRAN,UMM3
UMM2:	.word	DROP
UMM3:	.word	RFROM
	.word	DONXT,UMM1
	.word	DROP,SWAP,EXIT
UMM4:	.word	DROP,DDROP
	.word	DOLIT,-1,DUPP,EXIT

;   M/MOD	( d n -- r q )
;		Signed floored divide of double by single. Return mod and quotient.

	COLON	5,"M/MOD",MSMOD
	.word	DUPP,ZLESS,DUPP,TOR
	.word	QBRAN,MMOD1
	.word	NEGAT,TOR,DNEGA,RFROM
MMOD1:	.word	TOR,DUPP,ZLESS
	.word	QBRAN,MMOD2
	.word	RAT,PLUS
MMOD2:	.word	RFROM,UMMOD,RFROM
	.word	QBRAN,MMOD3
	.word	SWAP,NEGAT,SWAP
MMOD3:	.word	EXIT

;   /MOD	( n n -- r q )
;		Signed divide. Return mod and quotient.

	COLON	4,"/MOD",SLMOD
	.word	OVER,ZLESS,SWAP,MSMOD,EXIT

;   MOD		( n n -- r )
;		Signed divide. Return mod only.

	COLON	3,"MOD",MODD
	.word	SLMOD,DROP,EXIT

;   /		( n n -- q )
;		Signed divide. Return quotient only.

	COLON	1,"/",SLASH
	.word	SLMOD,SWAP,DROP,EXIT
	

;; Multiply
	.ifdef hardwareUM
	.globl	hwmult
	CODE	3,"UM*",UMSTA
	mov	TOS,r15
	pop	r14
	call	#hwmult
	push	r14
	mov	r15,TOS
	next
	.else
;   UM*		( u u -- ud )
;		Unsigned multiply. Return double product.

	COLON	3,"UM*",UMSTA
	.word	DOLIT,0,SWAP,DOLIT,15,TOR
UMST1:	.word	DUPP,UPLUS,TOR,TOR
	.word	DUPP,UPLUS,RFROM,PLUS,RFROM
	.word	QBRAN,UMST2
	.word	TOR,OVER,UPLUS,RFROM,PLUS
UMST2:	.word	DONXT,UMST1
	.word	ROT,DROP,EXIT
	.endif
;   *		( n n -- n )
;		Signed multiply. Return single product.

	COLON	1,"*",STAR
	.word	UMSTA,DROP,EXIT

;   M*		( n n -- d )
;		Signed multiply. Return double product.

	COLON	2,"M*",MSTAR
	.word	DDUP,XORR,ZLESS,TOR
	.word	ABSS,SWAP,ABSS,UMSTA
	.word	RFROM
	.word	QBRAN,MSTA1
	.word	DNEGA
MSTA1:	.word	EXIT

;   */MOD	( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

	COLON	5,"*/MOD",SSMOD
	.word	TOR,MSTAR,RFROM,MSMOD,EXIT

;   */		( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

	COLON	2,"*/",STASL
	.word	SSMOD,SWAP,DROP,EXIT

;; Miscellaneous

;   CELL+	( a -- a )
;		Add cell size in byte to address.

	COLON  5,"CELL+",CELLP
	.word	DOLIT,CELLL,PLUS,EXIT

;   CELL-	( a -- a )
;		Subtract cell size in byte from address.

	COLON  5,"CELL-",CELLM
	.word	DOLIT,0-CELLL,PLUS,EXIT

;   CELLS	( n -- n )
;		Multiply tos by cell size in bytes.

	COLON  5,"CELLS",CELLS
	.word	DDUP,PLUS,EXIT

        ;; ALIGNED   ( b -- a )
        ;; Align the address with a CELL boundry
        ;; Cell size must be power of 2

        COLON  7, "ALIGNED",ALGND
        .word   DUPP, DOLIT,CELLL,ONEM,ANDD
        .word   DUPP,QBRAN,ALGND1
        .word   DOLIT,CELLL, SWAP,SUBB
ALGND1: .word   PLUS,EXIT
	
;   1+          ( a -- a )
;		Add 1 to tos

	COLON  2,"1+",ONEP
	.word      DOLIT,1,PLUS,EXIT

;   1-          ( a -- a )
;		Subtract 1 from tos
	CODE  2,"1-",ONEM
	dec	TOS
	next
	;; 	COLON  2,"1-",ONEM
	;; 	.word      DOLIT,-1,PLUS,EXIT

;   2/          ( n -- n )
;		Divide tos by 2

	COLON  2,"2/",TWOSL
	.word      DOLIT,2,SLASH,EXIT

;   BL		( -- 32 )
;		Return 32, the blank character.

	COLON	2,"BL",BLANK
	.word	DOLIT,' ',EXIT

;   >CHAR       ( c -- c )
;		Filter non-printing characters.

	COLON  5,">CHAR",TCHAR
	.word	DOLIT,0x07F,ANDD,DUPP;mask msb
	.word	DOLIT,127,BLANK,WITHI;check for printable
	.word	QBRAN,TCHA1
	.word	DROP,DOLIT,'_'	;replace non-printables
TCHA1:	.word	EXIT

;   DEPTH	( -- n )
;		Return the depth of the data stack.

	COLON	5,"DEPTH",DEPTH
	.word      SPAT,DOLIT,SPP,SWAP,SUBB
	.word	DOLIT,CELLL,SLASH,EXIT

;   PICK	( ... +n -- ... w )
;		Copy the nth stack item to tos.
	;; Modified to handle TOS being in a register
	;; 	COLON	4,"PICK",PICK
	;; 	.word	QDUP,QBRAN,PICK1
	;; 	.word   ONEM,CELLS
	;; 	.word	SPAT,PLUS,AT,EXIT
	;; PICK1:	.word	DDUP,EXIT
	;; I can't seem to make the high level version work, but the code
	;; is simple. Double TOS and add to the PSP.
	CODE	4,"PICK",PICK
	add	TOS,TOS
	add	PSP,TOS
	mov	@(TOS),TOS
	next

;; Memory access

;   +!		( n a -- )
;		Add n to the contents at address a.

	COLON	2,"+!",PSTOR
	.word	SWAP,OVER,AT,PLUS
	.word	SWAP,STORE,EXIT

;   2!		( d a -- )
;		Store the double integer to address a.

	COLON	2,"2!",DSTOR
	.word	SWAP,OVER,STORE
	.word	CELLP,STORE,EXIT

;   2@		( a -- d )
;		Fetch double integer from address a.

	COLON	2,"2@",DAT
	.word	DUPP,CELLP,AT
	.word	SWAP,AT,EXIT

;   COUNT	( b -- b +n )
;		Return count byte of a string and add 1 to byte address.

	COLON	5,"COUNT",COUNT
	.word      DUPP,ONEP
	.word	SWAP,CAT,EXIT

;   HERE	( -- a )
;		Return the top of the code dictionary.

	COLON	4,"HERE",HERE
	.word	CP,AT,EXIT

;   PAD		( -- a )
;		Return the address of the text buffer above the code dictionary.

	COLON	3,"PAD",PAD
	.word	HERE,DOLIT,80,PLUS,EXIT

;   TIB		( -- a )
;		Return the address of the terminal input buffer.

	COLON	3,"TIB",TIB
	.word	NTIB,CELLP,AT,EXIT

;   @EXECUTE	( a -- )
;		Execute vector stored in address a.

	COLON	8,"@EXECUTE",ATEXE
	.word	AT,QDUP		;?address or zero
	.word	QBRAN,EXE1
	.word	EXECU		;execute if non-zero
EXE1:	.word	EXIT		;do nothing if zero

;   CMOVE	( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

	COLON	5,"CMOVE",CMOVE
	.word	TOR
	.word	BRAN,CMOV2
CMOV1:	.word	TOR,DUPP,CAT
	.word	RAT,CSTOR
	.word      ONEP
	.word      RFROM,ONEP
CMOV2:		.word	DONXT,CMOV1
	.word	DDROP,EXIT

;   FILL	( b u c -- )
;		Fill u bytes of character c to area beginning at b.

	COLON	4,"FILL",FILL
	.word	SWAP,TOR,SWAP
	.word	BRAN,FILL2
FILL1:  .word      DDUP,CSTOR,ONEP
FILL2:	.word	DONXT,FILL1
	.word	DDROP,EXIT

;   ERASE       ( b u -- )
;               Erase u bytes beginning at b.
	#undef ERASE			
	COLON  5,"ERASE",ERASE
	.word      DOLIT,0,FILL
	.word      EXIT

;   PACK$	( b u a -- a )
;		Build a counted string with u characters from b. Null fill.

	COLON  5,"PACK$",PACKS
	;; 	.word      DUPP,TOR          ;strings only on cell boundary
	;; 	.word      DDUP,CSTOR,ONEP ;save count
	;; 	.word	SWAP,CMOVE,RFROM,EXIT;move string
	
	;;  Version 1 PACK$
        .word      ALGND,DUPP,TOR          ;strings only on cell boundary
        .word      OVER,DUPP,DOLIT,0
        .word      DOLIT,CELLL,UMMOD,DROP  ;count mod cell
        .word      SUBB,OVER,PLUS
        .word      DOLIT,0,SWAP,STORE      ;null fill cell
        .word      DDUP,CSTOR,DOLIT,1,PLUS ;save count
        .word      SWAP,CMOVE,RFROM,EXIT   ;move string
	
;; Numeric output, single precision

;   DIGIT	( u -- c )
;		Convert digit u to a character.

	COLON  5,"DIGIT",DIGIT
	.word	DOLIT,9,OVER,LESS
	.word	DOLIT,7,ANDD,PLUS
	.word	DOLIT,'0',PLUS,EXIT

;   EXTRACT	( n base -- n c )
;		Extract the least significant digit from n.

	        COLON  7,"EXTRACT",EXTRC
	.word	DOLIT,0,SWAP,UMMOD
	.word	SWAP,DIGIT,EXIT

;   <#		( -- )
;		Initiate the numeric output process.

	COLON	2,"<#",BDIGS
	.word	PAD,HLD,STORE,EXIT

;   HOLD	( c -- )
;		Insert a character into the numeric output string.

	COLON	4,"HOLD",HOLD
	        .word      HLD,AT,ONEM
	.word	DUPP,HLD,STORE,CSTOR,EXIT

;   #		( u -- u )
;		Extract one digit from u and append the digit to output string.

	COLON	1,"#",DIG
	.word	BASE,AT,EXTRC,HOLD,EXIT

;   #S		( u -- 0 )
;		Convert u until all digits are added to the output string.

	COLON	2,"#S",DIGS
DIGS1:	.word	DIG,DUPP
	.word	QBRAN,DIGS2
	.word	BRAN,DIGS1
DIGS2:	.word	EXIT

;   SIGN	( n -- )
;		Add a minus sign to the numeric output string.

	COLON	4,"SIGN",SIGN
	.word	ZLESS
	.word	QBRAN,SIGN1
	.word	DOLIT,'-',HOLD
SIGN1:	.word	EXIT

;   #>		( w -- b u )
;		Prepare the output string to be TYPE"d.

	COLON	2,"#>",EDIGS
	.word	DROP,HLD,AT
	.word	PAD,OVER,SUBB,EXIT

;   str		( w -- b u )
;		Convert a signed integer to a numeric string.

	COLON  3,"str",STR
	.word	DUPP,TOR,ABSS
	.word	BDIGS,DIGS,RFROM
	.word	SIGN,EDIGS,EXIT

;   HEX		( -- )
;		Use radix 16 as base for numeric conversions.

	COLON	3,"HEX",HEX
	.word	DOLIT,16,BASE,STORE,EXIT

;   DECIMAL	( -- )
;		Use radix 10 as base for numeric conversions.

	COLON	7,"DECIMAL",DECIM
	.word	DOLIT,10,BASE,STORE,EXIT

;; Numeric input, single precision

;   DIGIT?	( c base -- u t )
;		Convert a character to its numeric value. A flag indicates success.

	COLON  7,"DIGIT?",DIGTQ
	.word	TOR,DOLIT,'0',SUBB
	.word	DOLIT,9,OVER,LESS
	.word	QBRAN,DGTQ1
	.word	DOLIT,7,SUBB
	.word	DUPP,DOLIT,10,LESS,ORR
DGTQ1:	.word	DUPP,RFROM,ULESS,EXIT

;   NUMBER?	( a -- n T | a F )
;		Convert a number string to integer. Push a flag on tos.

	COLON	7,"NUMBER?",NUMBQ
	.word	BASE,AT,TOR,DOLIT,0,OVER,COUNT
	.word	OVER,CAT,DOLIT,'$',EQUAL
	.word	QBRAN,NUMQ1
	.word      HEX,SWAP,ONEP
	.word      SWAP,ONEM
NUMQ1:	.word	OVER,CAT,DOLIT,'-',EQUAL,TOR
	.word	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
	.word	QBRAN,NUMQ6
	.word      ONEM,TOR
NUMQ2:	.word	DUPP,TOR,CAT,BASE,AT,DIGTQ
	.word	QBRAN,NUMQ4
	.word	SWAP,BASE,AT,STAR,PLUS,RFROM
	.word      ONEP
	.word	DONXT,NUMQ2
	.word	RAT,SWAP,DROP
	.word	QBRAN,NUMQ3
	.word	NEGAT
NUMQ3:	.word	SWAP
	.word	BRAN,NUMQ5
NUMQ4:	.word	RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:	.word	DUPP
NUMQ6:	.word	RFROM,DDROP
	.word	RFROM,BASE,STORE,EXIT

;; Basic I/O

;   KEY		( -- c )
;		Wait for and return an input character.

	COLON	3,"KEY",KEY
KEY1:	.word	QKEY
	.word	QBRAN,KEY1
	.word	EXIT

;   NUF?	( -- t )
;		Return false if no input, else pause and if CR return true.

	COLON	4,"NUF?",NUFQ
	.word	QKEY,DUPP
	.word	QBRAN,NUFQ1
	.word	DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:	.word	EXIT

;   SPACE	( -- )
;		Send the blank character to the output device.

	COLON	5,"SPACE",SPACE
	.word	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;		Send n spaces to the output device.

	COLON	6,"SPACES",SPACS
	.word	DOLIT,0,MAX,TOR
	.word	BRAN,CHAR2
CHAR1:	.word	SPACE
CHAR2:	.word	DONXT,CHAR1
	.word	EXIT

;   TYPE	( b u -- )
;		Output u characters from b.

	COLON	4,"TYPE",TYPES
	.word	TOR
	.word	BRAN,TYPE2
TYPE1:	.word	DUPP,CAT,EMIT
	.word      ONEP
TYPE2:	.word	DONXT,TYPE1
	.word	DROP,EXIT

;   CR		( -- )
;		Output a carriage return and a line feed.

	COLON	2,"CR",CR
	.word	DOLIT,CRR,EMIT
	.word	DOLIT,LF,EMIT,EXIT

;   do$		( -- a )
;		Return the address of a compiled string.

	COLON  COMPO+3,"do$",DOSTR
	.word	RFROM,RAT,RFROM,COUNT,PLUS
;               .word      TOR,SWAP,TOR,EXIT
	.word      TOR,SWAP,TOR,EXIT

;   $"|		( -- a )
;		Run time routine compiled by $". Return address of a compiled string.

	COLON  COMPO+3,"$\"|",STRQP
	.word	DOSTR,EXIT	;force a call to do$

;   ."|		( -- )
;		Run time routine of ." . Output a compiled string.

	COLON  COMPO+3,".\"|",DOTQP
	.word	DOSTR,COUNT,TYPES,EXIT

;   .R		( n +n -- )
;		Display an integer in a field of n columns, right justified.

	COLON	2,".R",DOTR
	.word	TOR,STR,RFROM,OVER,SUBB
	.word	SPACS,TYPES,EXIT

;   U.R		( u +n -- )
;		Display an unsigned integer in n column, right justified.

	COLON	3,"U.R",UDOTR
	.word	TOR,BDIGS,DIGS,EDIGS
	.word	RFROM,OVER,SUBB
	.word	SPACS,TYPES,EXIT

;   U.		( u -- )
;		Display an unsigned integer in free format.

	COLON	2,"U.",UDOT
	.word	BDIGS,DIGS,EDIGS
	.word	SPACE,TYPES,EXIT

;   .		( w -- )
;		Display an integer in free format, preceeded by a space.

	COLON	1,".",DOT
	.word	BASE,AT,DOLIT,10,XORR;?decimal
	.word	QBRAN,DOT1
	.word	UDOT,EXIT	;no, display unsigned
DOT1:	.word	STR,SPACE,TYPES,EXIT;yes, display signed

;   ?		( a -- )
;		Display the contents in a memory cell.

	COLON	1,"?",QUEST
	.word	AT,DOT,EXIT

;; Parsing

;   parse	( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

	COLON  5,"parse",PARS
	.word	TEMP,STORE,OVER,TOR,DUPP
	.word	QBRAN,PARS8
	.word      ONEM,TEMP,AT,BLANK,EQUAL
	.word	QBRAN,PARS3
	.word	TOR
PARS1:	.word	BLANK,OVER,CAT	;skip leading blanks ONLY
	.word	SUBB,ZLESS,INVER
	.word	QBRAN,PARS2
	.word      ONEP
	.word	DONXT,PARS1
	.word	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:	.word	RFROM
PARS3:	.word	OVER,SWAP
	.word	TOR
PARS4:	.word	TEMP,AT,OVER,CAT,SUBB;scan for delimiter
	.word	TEMP,AT,BLANK,EQUAL
	.word	QBRAN,PARS5
	.word	ZLESS
PARS5:	.word	QBRAN,PARS6
	.word      ONEP
	.word	DONXT,PARS4
	.word	DUPP,TOR
	.word	BRAN,PARS7
PARS6:	.word	RFROM,DROP,DUPP
	.word      ONEP,TOR
PARS7:	.word	OVER,SUBB
	.word	RFROM,RFROM,SUBB,EXIT
PARS8:	.word	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

	COLON	5,"PARSE",PARSE
	.word	TOR,TIB,INN,AT,PLUS;current input buffer pointer
	.word	NTIB,AT,INN,AT,SUBB;remaining count
	.word	RFROM,PARS,INN,PSTOR,EXIT

;   .(		( -- )
;		Output following string up to next ) .

	COLON	IMEDD+2,".(",DOTPR
	.word	DOLIT,')',PARSE,TYPES,EXIT

;   (		( -- )
;		Ignore following string up to next ) . A comment.

	COLON	IMEDD+1,"(",PAREN
	.word	DOLIT,')',PARSE,DDROP,EXIT

;   \		( -- )
;		Ignore following text till the end of line.

	COLON	IMEDD+1,"\\",BKSLA
	.word	NTIB,AT,INN,STORE,EXIT

;   WORD	( c -- a ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

	COLON	4,"WORD",WORDD
	.word   PARSE
	.word   HERE,CELLP
	.word	PACKS,EXIT

;   TOKEN	( -- a ; <string> )
;		Parse a word from input stream and copy it to name dictionary.

	COLON  5,"TOKEN",TOKEN
	.word      BLANK,WORDD,EXIT

;; Dictionary search

;   NAME>	( na -- ca )
;		Return a code address given a name address.
	;; Aligned please
	
	COLON  5,"NAME>",NAMET
	.word      COUNT,DOLIT,31,ANDD
	.word      PLUS,ALGND,EXIT

;   SAME?	( a a u -- a a f \ -0+ )
;		Compare u chars in two strings. Return 0 if identical.

	COLON  6,"SAME?",SAMEQ
	.word      ONEM,TOR
	.word	BRAN,SAME2
SAME1:  .word      OVER,RAT,PLUS,CAT
	.word      OVER,RAT,PLUS,CAT
	.word	SUBB,QDUP
	.word	QBRAN,SAME2
	.word	RFROM,DROP,EXIT
SAME2:	.word	DONXT,SAME1
	.word	DOLIT,0,EXIT

;   find	( a va -- ca na | a F )
;		Search a vocabulary for a string. Return ca and na if succeeded.

	COLON  4,"find",FIND
	.word	SWAP,DUPP,CAT
	.word      TEMP,STORE
	.word	DUPP,AT,TOR,CELLP,SWAP
FIND1:	.word	AT,DUPP
	.word	QBRAN,FIND6
	.word	DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
	.word	QBRAN,FIND2
	.word	CELLP,DOLIT,-1
	.word	BRAN,FIND3
FIND2:	.word	CELLP,TEMP,AT,SAMEQ
FIND3:	.word	BRAN,FIND4
FIND6:	.word	RFROM,DROP
	.word	SWAP,CELLM,SWAP,EXIT
FIND4:	.word	QBRAN,FIND5
	.word	CELLM,CELLM
	.word	BRAN,FIND1
FIND5:	.word	RFROM,DROP,SWAP,DROP
	.word	CELLM
	.word	DUPP,NAMET,SWAP,EXIT

;   NAME?	( a -- ca na | a F )
;		Search all context vocabularies for a string.

	COLON  5,"NAME?",NAMEQ
	.word      CNTXT,FIND,EXIT

;; Terminal response

;   ^H		( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

	COLON  2,"^h",BKSP
	.word	TOR,OVER,RFROM,SWAP,OVER,XORR
	.word	QBRAN,BACK1
	.word      DOLIT,BKSPP,EMIT,ONEM
	.word      BLANK,EMIT
	.word      DOLIT,BKSPP,EMIT
BACK1:	.word	EXIT

;   TAP		( bot eot cur c -- bot eot cur )
;		Accept and echo the key stroke and bump the cursor.

	COLON  3,"TAP",TAP
	.word      DUPP,EMIT
	.word      OVER,CSTOR,ONEP,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.

	COLON  4,"kTAP",KTAP
	.word	DUPP,DOLIT,CRR,XORR
	.word	QBRAN,KTAP2
	.word	DOLIT,BKSPP,XORR
	.word	QBRAN,KTAP1
	.word	BLANK,TAP,EXIT
KTAP1:	.word	BKSP,EXIT
KTAP2:	.word	DROP,SWAP,DROP,DUPP,EXIT

;   accept	( b u -- b u )
;	Accept characters to input buffer. Return with actual count.

	COLON  6,"ACCEPT",ACCEP
	.word	OVER,PLUS,OVER
ACCP1:	.word	DDUP,XORR
	.word	QBRAN,ACCP4
	.word	KEY,DUPP
;		.word	BLANK,SUBB,DOLIT,95,ULESS
	.word	BLANK,DOLIT,127,WITHI
	.word	QBRAN,ACCP2
	.word	TAP
	.word	BRAN,ACCP3
ACCP2:  .word      KTAP
ACCP3:	.word	BRAN,ACCP1
ACCP4:	.word	DROP,OVER,SUBB,EXIT

;   QUERY	( -- )
;		Accept input stream to terminal input buffer.

	COLON	5,"QUERY",QUERY
	.word      TIB,DOLIT,80,ACCEP,NTIB,STORE
	.word	DROP,DOLIT,0,INN,STORE,EXIT

;   ABORT	( -- )
;		Reset data stack and jump to QUIT.

	COLON	5,"ABORT",ABORT
	.word      PRESE,QUIT

;   abort"	( f -- )
;		Run time routine of ABORT" . Abort with a message.

	COLON  COMPO+6,"abort\"",ABORQ
	.word      QBRAN,ABOR2             ;text flag
	.word      DOSTR
ABOR1:  .word      SPACE,COUNT,TYPES
	.word      DOLIT,'?',EMIT,CR,ABORT  ;pass error string
ABOR2:  .word      DOSTR,DROP,EXIT         ;drop error

;; The text interpreter

;   $INTERPRET	( a -- )
;		Interpret a word. If failed, try to convert it to an integer.

	COLON  10,"$INTERPRET",INTER
	.word	NAMEQ,QDUP	;?defined
	.word	QBRAN,INTE1
	.word	CAT,DOLIT,COMPO,ANDD;?compile only lexicon bits
	.word	ABORQ
	.byte	13
	.ascii	" compile only"
	.word	EXECU,EXIT	;execute defined word
INTE1:	.word      NUMBQ            ;convert a number
	.word      QBRAN,ABOR1
	.word	EXIT

;   [		( -- )
;		Start the text interpreter.

	COLON	IMEDD+1,"[",LBRAC
	.word	DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK		( -- )
;		Display "ok" only while interpreting.

	COLON  3,".OK",DOTOK
	.word	DOLIT,INTER,TEVAL,AT,EQUAL
	.word	QBRAN,DOTO1
	.word	DOTQP
	.byte	3
	.ascii	" ok"
DOTO1:	.word	CR,EXIT

;   ?STACK	( -- )
;		Abort if the data stack underflows.

	COLON  6,"?STACK",QSTAC
	.word	DEPTH,ZLESS	;check only for underflow
	.word	ABORQ
	.byte      11
	.ascii " underflow "
	.word	EXIT

;   EVAL	( -- )
;		Interpret the input stream.

	COLON	4,"EVAL",EVAL
EVAL1:	.word	TOKEN,DUPP,CAT	;?input stream empty
	.word	QBRAN,EVAL2
	.word	TEVAL,ATEXE,QSTAC;evaluate input, check stack
	.word	BRAN,EVAL1
EVAL2:  .word      DROP,DOTOK,EXIT   ;prompt

;   PRESET	( -- )
;		Reset data stack pointer and the terminal input buffer.

	COLON  6,"PRESET",PRESE
	.word      DOLIT,SPP,SPSTO
	.word	DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;   QUIT	( -- )
;		Reset return stack pointer and start text interpreter.

	COLON	4,"QUIT",QUIT
	.word      DOLIT,RPP,RPSTO          ;reset return stack pointer
QUIT1:	.word	LBRAC		;start interpretation
QUIT2:	.word	QUERY		;get input
	.word      EVAL
	.word      BRAN,QUIT2             ;continue till error

;; The compiler

;   '		( -- ca )
;		Search context vocabularies for the next word in input stream.

	COLON	1,"'",TICK
	.word	TOKEN,NAMEQ	;?defined
	.word      QBRAN,ABOR1
	.word	EXIT		

;   ALLOT	( n -- )
;		Allocate n bytes to the code dictionary.

	COLON	5,"ALLOT",ALLOT
	.word	CP,PSTOR,EXIT	;adjust code pointer

;   ,		( w -- )
;		Compile an integer into the code dictionary.

	COLON	1,",",COMMA
	.word	HERE,DUPP,CELLP	;cell boundary
	.word	CP,STORE,STORE,EXIT;adjust code pointer and compile

;   [COMPILE]	( -- ; <string> )
;		Compile the next immediate word into code dictionary.

	COLON	IMEDD+9,"[COMPILE]",BCOMP
	.word	TICK,COMMA,EXIT

;   COMPILE	( -- )
;		Compile the next address in colon list to code dictionary.

	COLON	COMPO+7,"COMPILE",COMPI
	.word	RFROM,DUPP,AT,COMMA;compile address
	.word	CELLP,TOR,EXIT	;adjust return address

;   LITERAL	( w -- )
;		Compile tos to code dictionary as an integer literal.

	COLON	IMEDD+7,"LITERAL",LITER
	.word	COMPI,DOLIT,COMMA,EXIT

;   [']		( -- ca )
;		Search context vocabularies for the next word in input stream.
; 		Compile address as literal.

	COLON	IMEDD+3,"[']",BTICK
	.word	TOKEN,NAMEQ	;?defined
	.word      QBRAN,ABOR1
	.word	LITER,EXIT ; yes, compile address as literal

;   $,"		( -- )
;		Compile a literal string up to next " .

	COLON  3,"$,\"",STRCQ
	.word   DOLIT,'\"',PARSE,HERE,PACKS   ;string to code dictionary
	.word   COUNT,PLUS,ALGND        ;calculate aligned end of string
	.word	CP,STORE,EXIT	;adjust the code pointer

;; Structures

;   FOR		( -- a )
;		Start a FOR-NEXT loop structure in a colon definition.

	COLON	IMEDD+3,"FOR",FOR
	.word	COMPI,TOR,HERE,EXIT

;   BEGIN	( -- a )
;		Start an infinite or indefinite loop structure.

	COLON	IMEDD+5,"BEGIN",BEGIN
	.word	HERE,EXIT

;   NEXT	( a -- )
;		Terminate a FOR-NEXT loop structure.

	COLON	IMEDD+4,"NEXT",NEXT
	.word	COMPI,DONXT,COMMA,EXIT

;   UNTIL	( a -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

	COLON	IMEDD+5,"UNTIL",UNTIL
	.word	COMPI,QBRAN,COMMA,EXIT

;   AGAIN	( a -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

	COLON	IMEDD+5,"AGAIN",AGAIN
	.word	COMPI,BRAN,COMMA,EXIT

;   IF		( -- A )
;		Begin a conditional branch structure.

	COLON	IMEDD+2,"IF",IFF
	.word	COMPI,QBRAN,HERE
	.word	DOLIT,0,COMMA,EXIT

;   AHEAD	( -- A )
;		Compile a forward branch instruction.

	COLON	IMEDD+5,"AHEAD",AHEAD
	.word	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT	( A a -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

	COLON	IMEDD+6,"REPEAT",REPEA
	.word	AGAIN,HERE,SWAP,STORE,EXIT

;   THEN	( A -- )
;		Terminate a conditional branch structure.

		COLON	IMEDD+4,"THEN",THENN
	.word	HERE,SWAP,STORE,EXIT

;   AFT		( a -- a A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

	COLON	IMEDD+3,"AFT",AFT
	.word	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

	COLON	IMEDD+4,"ELSE",ELSEE
	.word	AHEAD,SWAP,THENN,EXIT

;   WHILE	( a -- A a )
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

	COLON	IMEDD+5,"WHILE",WHILE
	.word	IFF,SWAP,EXIT

;   ABORT"	( -- ; <string> )
;		Conditional abort with an error message.

	COLON	IMEDD+6,"ABORT\"",ABRTQ
	.word	COMPI,ABORQ,STRCQ,EXIT

;   $"		( -- ; <string> )
;		Compile an inline string literal.

	COLON	IMEDD+2,"$\"",STRQ
	.word	COMPI,STRQP,STRCQ,EXIT

;   ."		( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

	COLON	IMEDD+2,".\"",DOTQ
	.word	COMPI,DOTQP,STRCQ,EXIT

;; Name compiler

;   ?UNIQUE	( a -- a )
;		Display a warning message if the word already exists.

	COLON  7,"?UNIQUE",UNIQU
	.word	DUPP,NAMEQ	;?name exists
	.word	QBRAN,UNIQ1
	.word	DOTQP		;redefinitions are OK
	.byte	7
	.ascii " reDef "	;but the user should be warned
	.word	OVER,COUNT,TYPES;just in case its not planned
UNIQ1:	.word	DROP,EXIT

;   $,n		( na -- )
;		Build a new dictionary name using the string at na.

	COLON  3,"$,n",SNAME
	.word	DUPP,CAT	;?null input
	.word	QBRAN,PNAM1
	.word	UNIQU		;?redefinition
	.word      DUPP,COUNT,PLUS,ALGND
	.word      CP,STORE
	.word	DUPP,LAST,STORE	;save na for vocabulary link
	.word	CELLM		;link address
	.word      CNTXT,AT,SWAP
	.word	STORE,EXIT	;save code pointer
PNAM1:  .word      STRQP
	.byte	5
	.ascii	" name"	;null input
	.word      BRAN,ABOR1

;; FORTH compiler

;   $COMPILE	( a -- )
;		Compile next word to code dictionary as a token or literal.

	COLON  8,"$COMPILE",SCOMP
	.word	NAMEQ,QDUP	;?defined
	.word	QBRAN,SCOM2
	.word	AT,DOLIT,IMEDD,ANDD;?immediate
	.word	QBRAN,SCOM1
	.word	EXECU,EXIT	;its immediate, execute
SCOM1:	.word	COMMA,EXIT	;its not immediate, compile
SCOM2:  .word      NUMBQ            ;try to convert to number
	.word      QBRAN,ABOR1
	.word	LITER,EXIT	;compile number as integer

;   OVERT	( -- )
;		Link a new word into the current vocabulary.

	COLON  5,"OVERT",OVERT
	.word      LAST,AT,CNTXT,STORE,EXIT

;   ;		( -- )
;		Terminate a colon definition.

	COLON	IMEDD+COMPO+1,";",SEMIS
	.word	COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]		( -- )
;		Start compiling the words in the input stream.

	COLON	1,"]",RBRAC
	.word	DOLIT,SCOMP,TEVAL,STORE,EXIT

;   call,	( ca -- )
;		Assemble a call instruction to ca.

	COLON  5,"call,",CALLC
	.word	DOLIT,CALLL,COMMA  ;DTC msp430 call (call r7)
	.word	EXIT

;   :		( -- ; <string> )
;		Start a new colon definition using next word as its name.

	COLON	1,":",COLON
	.word	TOKEN,SNAME
	.word	CALLC,RBRAC,EXIT

;   IMMEDIATE	( -- )
;		Make the last compiled word an immediate word.

	COLON	9,"IMMEDIATE",IMMED
	.word	DOLIT,IMEDD,LAST,AT,AT,ORR
	.word	LAST,AT,STORE,EXIT

;; Defining words

;   CREATE	( -- ; <string> )
;		Compile a new array entry without allocating code space.

	COLON	6,"CREATE",CREAT
	.word	TOKEN,SNAME,OVERT
	.word	CALLC
	.word	DOLIT,DOVAR,COMMA,EXIT

;   VARIABLE	( -- ; <string> )
;		Compile a new variable initialized to 0.

	COLON	8,"VARIABLE",VARIA
	.word	CREAT,COMPI,0,EXIT

; <BUILDS	( -- )
;		Compile a new header that calls dodoes
	
	COLON	7,"<BUILDS",BUILDS
	.word	TOKEN,SNAME,OVERT ; create header
	.word	COMPI,CALLD	  ; call dodoes
	.word	COMPI,0,EXIT	  ; reserve space for use by dodoes

; DOES>		( -- )
;		Set address of DOES> part of definition in parameter field
	
	COLON	5,"DOES>",DOES
	.word	RFROM		; get parameter field address
	.word	LAST,AT,NAMET	; and address of code field
	.word	CELLP,STORE	; store address of DOES> part of code
	.word	EXIT
	
;; Tools

;   _TYPE	( b u -- )
;		Display a string. Filter non-printing characters.

	COLON  5,"_TYPE",UTYPE
	.word	TOR		;start count down loop
	.word	BRAN,UTYP2	;skip first pass
UTYP1:	.word	DUPP,CAT,TCHAR,EMIT;display only printable
	.word      ONEP            ;increment address
UTYP2:	.word	DONXT,UTYP1	;loop till done
	.word	DROP,EXIT

;   dm+		( a u -- a )
;		Dump u bytes from , leaving a+u on the stack.

	COLON  3,"dm+",DUMPP
	.word	OVER,DOLIT,4,UDOTR;display address
	.word	SPACE,TOR	;start count down loop
	.word	BRAN,PDUM2	;skip first pass
PDUM1:	.word	DUPP,CAT,DOLIT,3,UDOTR;display numeric data
	.word      ONEP            ;increment address
PDUM2:	.word	DONXT,PDUM1	;loop till done
	.word	EXIT

;   DUMP	( a u -- )
;		Dump u bytes from a, in a formatted manner.

	COLON	4,"DUMP",DUMP
	.word	BASE,AT,TOR,HEX	;save radix, set hex
	.word	DOLIT,16,SLASH	;change count to lines
	.word	TOR		;start count down loop
DUMP1:	.word	CR,DOLIT,16,DDUP,DUMPP;display numeric
	.word	ROT,ROT
	.word	DOLIT,2,SPACS,UTYPE;display printable characters
	.word	NUFQ,INVER	;user control
	.word	QBRAN,DUMP2
	.word	DONXT,DUMP1	;loop till done
	.word	BRAN,DUMP3
DUMP2:	.word	RFROM,DROP	;cleanup loop stack, early exit
DUMP3:	.word	DROP,RFROM,BASE,STORE;restore radix
	.word	EXIT

;   .S		( ... -- ... )
;		Display the contents of the data stack.

	COLON	2,".S",DOTS
	.word	CR,DEPTH	;stack depth
	.word	TOR		;start count down loop
	.word	BRAN,DOTS2	;skip first pass
DOTS1:	.word	RAT,PICK,DOT	;index stack, display contents
DOTS2:	.word	DONXT,DOTS1	;loop till done
	.word	DOTQP
	.byte      5
	.ascii	" <sp "
	.word	EXIT

;   >NAME	( ca -- na | F )
;		Convert code address to a name address.

	COLON  5,">NAME",TNAME
	.word   CNTXT                   ;vocabulary link
TNAM2:	.word	AT,DUPP		;?last word in a vocabulary
	.word   QBRAN,TNAM4
	.word	DDUP,NAMET,XORR	;compare
	.word	QBRAN,TNAM3
	.word	CELLM		;continue with next word
	.word	BRAN,TNAM2
TNAM3:  .word   SWAP,DROP,EXIT
TNAM4:  .word   DDROP,DOLIT,0,EXIT

;   .ID		( na -- )
;		Display the name at address.

	COLON  3,".ID",DOTID
	.word	QDUP		;if zero no name
	.word	QBRAN,DOTI1
	.word	COUNT,DOLIT,0x01F,ANDD;mask lexicon bits
	.word	UTYPE,EXIT	;display name string
DOTI1:	.word	DOTQP
	.byte	9
	.ascii	" {noName}"
	.word	EXIT

;   SEE		( -- ; <string> )
;               A simple decompiler. Updated for byte machines, 08mar98cht

	COLON	3,"SEE",SEE
	.word	TICK		;starting address
	.word   CR
SEE1:   .word   CELLP,DUPP,AT,DUPP      ;?does it contain a zero
	.word	QBRAN,SEE2
	.word	TNAME		;?is it a name
SEE2:	.word	QDUP		;name address or zero
	.word	QBRAN,SEE3
	.word	SPACE,DOTID	;display name
	.word   ONEP
	.word	BRAN,SEE4
SEE3:   .word   DUPP,CAT,UDOT            ;display number
SEE4:	.word	NUFQ		;user control
	.word	QBRAN,SEE1
	.word	DROP,EXIT

;   WORDS	( -- )
;		Display the names in the context vocabulary.

	COLON	5,"WORDS",WORDS
	.word   CR,CNTXT             ;only in context
WORS1:	.word	AT,QDUP		;?at end of list
	.word	QBRAN,WORS2
	.word	DUPP,SPACE,DOTID;display a name
	.word	CELLM,NUFQ	;user control
	.word	QBRAN,WORS1
	.word	DROP
WORS2:	.word	EXIT

;; Hardware reset

;   hi		( -- )
;		Display the sign-on message of eForth.

	COLON	2,"hi",HI
	.word      CR,DOTQP          ;initialize I/O
	.byte      19
	.ascii	"MSP430eForth  v"
	.byte	VER+'0','.',EXT+'0,' '     ;version
	.word	CR,EXIT

;   "BOOT       ( -- a )
;               The application startup vector.

	COLON  5,"\"BOOT",TBOOT
	.word      DOVAR
	.word      HI                      ;application to boot

;   COLD	( -- )
;		The hilevel cold start sequence.

	COLON	4,"COLD",COLD
COLD1:	.word	DOLIT,UZERO,DOLIT,UPP
	.word	DOLIT,ULAST-UZERO,CMOVE;initialize user area
	.word	PRESE		;initialize data stack and TIB
	.word   TBOOT,ATEXE             ;application boot
	.word   OVERT
	.word	QUIT		;start interpretation
	.word	BRAN,COLD1	;just in case

;===============================================================

	.equ	LASTN, _link       ;last name address in name dictionary

	.word	0		; padding in case there is an instruction fetch
	.align	10		; here. Align on 1K MPU boundary
	.equ	CTOP, .

	.end
