; -*-asm-*-    set emacs asm-mode
#include <msp430fr5969.h>
	;; All of the device specific startup stuff. The main program
	;; shouldn't even have to include msp430.h
	;;
	;; Conversely, this code doesn't know much about the FORTH
	;; implementation.
	;;
	;; Assumes that a MSP430FR5969 Launchpad is being used. Uses
	;; the backchannel UART for the console. The eFORTH core is
	;; protected from being over-written using the MPU. Nothing should
	;; ever need to write there.
	;;
	;; Support for use of the hardware multiplier is included.
	;; 
	;; After a power up reset execution starts at sys_init. When it is
	;; done it calls "main" which should be the entry point for the
	;; FORTH interpreter.
	;;
	;; Simple polled serial I/O is provided by the "rx" and "tx" routines.
	;; 
	.text
	.global main
	.func sys_init
sys_init:
	mov	#__stack,r1	       ; temp stack
	mov.w	#WDTPW|WDTHOLD,&WDTCTL ;disable watchdog

	;; Configure all I/O ports to outputs with low out
	;;
	mov.b	#0,&P1OUT
	mov.b	#0,&P2OUT
	mov.b	#0,&P3OUT
	mov.b	#0,&P4OUT
	mov	#0,&PJOUT

	mov.b	#-1,&P1DIR
	mov.b	#-1,&P2DIR
	mov.b	#-1,&P3DIR
	mov.b	#-1,&P4DIR
	mov	#-1,&PJDIR

	bic	#LOCKLPM5,&PM5CTL0	 ; Disable GPIO power on default

	mov	#CSKEY,&CSCTL0 ; unlock CS registers
	;;  Set ACLK = XT1; MCLK = DCO	
        mov.w   #SELA__LFXTCLK+SELS__DCOCLK+SELM__DCOCLK,&CSCTL2
	;; DCO defaults to 8MHz and that is good enough for now
	mov	#DIVA__1|DIVM__1|DIVS__8,&CSCTL3	    ; MCLK = DCO/1
	
	;; Startup LFXT using the 32768Hz crystal. This is used by the serial
	;; port to set bit rates.
	bis	#BIT4|BIT5,&PJSEL0
	bic	#LFXTOFF,&CSCTL4 ; turn on LFXT
	;; Wait for oscillator to start
oscwait:
	bic	#LFXTOFFG,&CSCTL5 ; clear LFXT fault flag
	bic	#OFIFG,&SFRIFG1
	bit	#OFIFG,&SFRIFG1	; test oscillator fault flag
	jnz	oscwait

	clr.b	&CSCTL0		; a word mode write locks things up!
	;;
	;; The timer isn't really used for anything yet. But it was
	;; a handy way to make a blinken light at one point.
	;; 
	mov.w	#TASSEL__ACLK|MC__CONTINUOUS|TACLR,&TA0CTL

	;; Serial port setup
	bis	#UCSWRST,&UCA0CTLW0	; reset UART
	;;  Use ACLK and set for 9600 bps
	mov	#(UCSSEL__ACLK|UCSWRST),&UCA0CTLW0	; use ACLK, 8N1
	mov	#3,&UCA0BRW	; set divider
	mov	#0x9200,&UCA0MCTLW
	bis.b	#3,&P2SEL1	; assign pins to TXD and RXD
	bic.b	#3,&P2SEL0

	bic	#UCSWRST,&UCA0CTLW0	; bring UART out of reset

	;; Configure the MPU to protect the code.
	mov	ULAST-4,r9	; get value of CTOP
	clrc
	rrc	r9		; align as required
	rra	r9
	rra	r9
	rra	r9

	mov	#MPUPW,&MPUCTL0	; write password to enable access to MPU
	mov	r9,&MPUSEGB1	; B1 = CTOP
	mov	#0x1000,&MPUSEGB2 ; B2 = 0x10000 (segment 3 is upper mem)
	bic	#MPUSEG1WE,&MPUSAM ; write protect segment 1
	bic	#MPUSEG1VS,&MPUSAM ; NMI on write violation
	mov   	#MPUPW+MPUENA+MPUSEGIE,&MPUCTL0 ; Enable MPU

	mov.b	#0,&MPUCTL0_H	; disable MPU access

	;; 	bis	#VMAIE,&SFRIE1	; enable VMA interrupt
	
	br	#main
	.endfunc
	
	;; Serial receive using UCA0. If no character is available, r14 is
	;; cleared. If a character is available r14 is set to all ones and
	;; the received character is placed in r15
	.func
	.global rx
rx:	nop
	clr	r15		; clear flag
	bit	#UCRXIFG,&UCA0IFG	; check char received flag
	jeq	rx_done
	mov	#-1,r15		; set flag
	mov	&UCA0RXBUF,r14	; get char
rx_done:
	ret
	
	.endfunc

	;; Simple polled serial transmit.
	;;
	;; The character in r15 is transmitted. The hardware is polled to see
	;; if it is ready first. This could hang forever.
	.func
	.global tx
tx:	bit	#UCTXIFG,&UCA0IFG	; see if ready for data
	jeq	tx
	mov	r15,&UCA0TXBUF	; send data
	ret
	.endfunc

	.globl phex
	.func phex
ptable:	.byte	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37
	.byte	0x38,0x39,0x41,0x42,0x43,0x44,0x45,0x46
;; Print the 16 bit value in r12 as hex
phex:	mov	r12,r13
	swpb	r13
	call	#hexbyte
	mov	r12,r13
	call	#hexbyte
	ret
	
hexbyte:
	mov	r13,r14
	rra	r14
	rra	r14
	rra	r14
	rra	r14
	call	#hexit
	mov	r13,r14
	call	#hexit
	ret
;; print one hex nibble in r14
hexit:	and	#0xf,r14; mask off nibble
	mov.b	ptable(r14),r15
	call	#tx
	ret
	.endfunc
;; print a carriage return and line feed
	.globl crlf
	.func   crlf
crlf:	
	mov	#'\r',r15; CR
	call	#tx
	mov	#'\n',r15; LF
	call	#tx
	ret
	.endfunc
;; print the string pointed to be r13 using the count in r12
	.globl  pstring
	.func	pstring
pstring:	
	mov.b	@r13+,r15; get next char
	and	#0x7f,r15
	call	#tx	; and print it
	dec	r12
	jnz	pstring	; repeat
	ret
	.endfunc
	
	;;
	;; 16X16 multiply with 32 bit result
	;;
	;; Putting this here rather than in the main FORTH code is slower
	;; but it means that file doesn't have to include the machine header.
	;; Not a big deal I guess. In any case, this is still a lot faster
	;; than the high level colon definition.
	;; 
	
	.func hwmult
	.globl hwmult
hwmult:	
	mov	r14,&MPY
	mov	r15,&OP2
	mov	&RESLO,r14
	mov	&RESHI,r15
	ret
	.endfunc

sysnmi_irq:
	add.w   &SYSSNIV,r0 ; add offset to PC
        reti	            ; SYSSNIV_NONE:        No interrupt
        reti                ; SYSSNIV_SVS:         SVS low_power reset
        reti                ; SYSSNIV_UBDIFG:      FRAM Uncorrectable bit Error
        reti                ; SYSSNIV_ACCTEIFG:    Access time error
        reti                ; SYSSNIV_MPUSEGPIFG:  MPUSEGPIFG violation
        reti                ; SYSSNIV_MPUSEGIIFG:  MPUSEGIIFG violation
        jmp MEM0            ; SYSSNIV_MPUSEG1IFG:  MPUSEG1IFG violation
        reti                ; SYSSNIV_MPUSEG2IFG:  MPUSEG2IFG violation
        reti                ; SYSSNIV_MPUSEG3IFG:  MPUSEG3IFG violation
        jmp MEM1            ; SYSSNIV_VMAIFG:      VMAIFG
        reti                ; SYSSNIV_JMBINIFG:    JMBINIFG
        reti                ; SYSSNIV_JMBOUTIFG:   JMBOUTIFG
        reti                ; SYSSNIV_CBDIFG:      FRAM Correctable Bit error

	
	;; Vacant memory access error
MEM1:	mov	#vmess,r13
	mov	#vmlen,r12
	jmp	pmess
	
	;; Segment 1 violation
MEM0: 			    ; IFG cleared automagically
	mov	#merr,r13
	mov	#mlen,r12
pmess:	call	#pstring
	mov	2(r1),r12
	call	#phex		; print PC from stack
	mov	#imess,r13
	mov	#ilen,r12
	call	#pstring
	mov	r5,r12		; print IPTR
	call	#phex
	call	#crlf

 	br	#main		;cold start

	
merr:	.ascii	"\r\nWrite Protection error, PC="
	.equ	mlen, .-merr
imess:	.ascii	" IPTR="
	.equ	ilen, .-imess
vmess:	.ascii	"\r\nVacant memory access error, PC="
	.equ	vmlen, .-vmess
	;; Interrupt vectors and a default handler that does nothing
	
nullirq:
	reti
	.section .vectors, "ax",@progbits
	.space   0x4c, 0xff
	;; Must be very carefull here not to write anything to the lower
	;; part of this area. It is "reserved" but is used as a password
	;; to protect the FRAM. Writing to it will effectively brick the
	;; device so that it can no longer be programmed.
	;; Recoverable using BSL to do a mass erase.
	;; Filling with 0xff is supposed to be safe...
	
 	.word	nullirq		; 0xffcc - AES256
	.word	nullirq		; 0xffce - RTC
	.word	nullirq		; 0xffd0 - Port 4
	.word	nullirq		; 0xffd2 - Port 3
	.word	nullirq		; 0xffd4 - Timer3_A2 CC1, TA
	.word	nullirq		; 0xffd6 - Timer3_A2 CC0
	.word	nullirq		; 0xffd8 - Port 2
	.word	nullirq		; 0xffda - Timer2_A3 CC1, TA
	.word	nullirq		; 0xffdc - Timer2_A3 CC0
	.word	nullirq		; 0xffde - Port 1
	.word	nullirq		; 0xffe0 - Timer1_3 CC1, TA
	.word	nullirq		; 0xffe2 - Timer2_3 CC0
	.word	nullirq		; 0xffe4 - DMA
	.word	nullirq		; 0xffe6 - USCI A1 Rx/Tx
	.word	nullirq		; 0xffe8 - Timer0_A5 CC1-4, TA
	.word	nullirq		; 0xffea - Timer0_A5 CC0
	.word	nullirq		; 0xffec - ADC
	.word	nullirq		; 0xffee - USCI B0 Rx/Tx
	.word	nullirq		; 0xfff0 - USCI A0 Rx/Tx
	.word	nullirq		; 0xfff2 - Watchdog timer
	.word	nullirq		; 0xfff4 - Timer0_B3 CC1-2, TB
	.word	nullirq		; 0xfff6 - Timer0_B3 CC0
	.word	nullirq		; 0xfff8 - Comparator B
	.word	nullirq		; 0xfffa - User NMI
	.word	sysnmi_irq	; 0xfffc - System NMI
	.word	sys_init	; 0xfffe - reset

	.end
